@@ -1,18 +1,15 @@
 #!/usr/bin/env python3
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
-import fcntl
 import os
 import sys
 from collections import defaultdict
 from contextlib import contextmanager, suppress
-from typing import (
-    DefaultDict, Dict, Generator, List, Optional, Sequence, Tuple, TYPE_CHECKING
-)
+from typing import TYPE_CHECKING, DefaultDict, Dict, Generator, List, Optional, Sequence, Tuple
 
 import kitty.fast_data_types as fast_data_types
 
-from .constants import is_macos, kitty_base_dir, shell_path, terminfo_dir
+from .constants import handled_signals, is_freebsd, is_macos, kitten_exe, kitty_base_dir, shell_path, terminfo_dir
 from .types import run_once
 from .utils import log_error, which
 
@@ -25,11 +22,10 @@ if TYPE_CHECKING:
 
 
 if is_macos:
-    from kitty.fast_data_types import (
-        cmdline_of_process as cmdline_, cwd_of_process as _cwd,
-        environ_of_process as _environ_of_process,
-        process_group_map as _process_group_map
-    )
+    from kitty.fast_data_types import cmdline_of_process as cmdline_
+    from kitty.fast_data_types import cwd_of_process as _cwd
+    from kitty.fast_data_types import environ_of_process as _environ_of_process
+    from kitty.fast_data_types import process_group_map as _process_group_map
 
     def cwd_of_process(pid: int) -> str:
         return os.path.realpath(_cwd(pid))
@@ -40,17 +36,26 @@ if is_macos:
             ans[pgid].append(pid)
         return ans
 
-    def cmdline_of_process(pid: int) -> List[str]:
+    def cmdline_of_pid(pid: int) -> List[str]:
         return cmdline_(pid)
 else:
 
-    def cmdline_of_process(pid: int) -> List[str]:
+    def cmdline_of_pid(pid: int) -> List[str]:
         with open(f'/proc/{pid}/cmdline', 'rb') as f:
             return list(filter(None, f.read().decode('utf-8').split('\0')))
 
-    def cwd_of_process(pid: int) -> str:
-        ans = f'/proc/{pid}/cwd'
-        return os.path.realpath(ans)
+    if is_freebsd:
+        def cwd_of_process(pid: int) -> str:
+            import subprocess
+            cp = subprocess.run(['pwdx', str(pid)], capture_output=True)
+            if cp.returncode != 0:
+                raise ValueError(f'Failed to find cwd of process with pid: {pid}')
+            ans = cp.stdout.decode('utf-8', 'replace').split()[1]
+            return os.path.realpath(ans)
+    else:
+        def cwd_of_process(pid: int) -> str:
+            ans = f'/proc/{pid}/cwd'
+            return os.path.realpath(ans)
 
     def _environ_of_process(pid: int) -> str:
         with open(f'/proc/{pid}/environ', 'rb') as f:
@@ -131,19 +136,12 @@ def environ_of_process(pid: int) -> Dict[str, str]:
     return parse_environ_block(_environ_of_process(pid))
 
 
-def remove_cloexec(fd: int) -> None:
-    fcntl.fcntl(fd, fcntl.F_SETFD, fcntl.fcntl(fd, fcntl.F_GETFD) & ~fcntl.FD_CLOEXEC)
-
-
-def remove_blocking(fd: int) -> None:
-    os.set_blocking(fd, False)
-
-
 def process_env() -> Dict[str, str]:
     ans = dict(os.environ)
     ssl_env_var = getattr(sys, 'kitty_ssl_env_var', None)
     if ssl_env_var is not None:
         ans.pop(ssl_env_var, None)
+    ans.pop('XDG_ACTIVATION_TOKEN', None)
     return ans
 
 
@@ -164,9 +162,14 @@ def set_default_env(val: Optional[Dict[str, str]] = None) -> None:
     setattr(default_env, 'lc_ctype_set_by_user', has_lctype)
 
 
+def set_LANG_in_default_env(val: str) -> None:
+    default_env().setdefault('LANG', val)
+
+
 def openpty() -> Tuple[int, int]:
     master, slave = os.openpty()  # Note that master and slave are in blocking mode
-    remove_cloexec(slave)
+    os.set_inheritable(slave, True)
+    os.set_inheritable(master, False)
     fast_data_types.set_iutf8_fd(master, True)
     return master, slave
 
@@ -195,13 +198,15 @@ class Child:
         stdin: Optional[bytes] = None,
         env: Optional[Dict[str, str]] = None,
         cwd_from: Optional['CwdRequest'] = None,
-        allow_remote_control: bool = False
+        is_clone_launch: str = '',
+        add_listen_on_env_var: bool = True,
     ):
-        self.allow_remote_control = allow_remote_control
+        self.is_clone_launch = is_clone_launch
+        self.add_listen_on_env_var = add_listen_on_env_var
         self.argv = list(argv)
         if cwd_from:
             try:
-                cwd = cwd_from.modify_argv_for_launch_with_cwd(self.argv) or cwd
+                cwd = cwd_from.modify_argv_for_launch_with_cwd(self.argv, env) or cwd
             except Exception as err:
                 log_error(f'Failed to read cwd of {cwd_from} with error: {err}')
         else:
@@ -209,53 +214,69 @@ class Child:
         self.cwd = os.path.abspath(cwd)
         self.stdin = stdin
         self.env = env or {}
+        self.is_default_shell = bool(self.argv and self.argv[0] == shell_path)
+        self.should_run_via_run_shell_kitten = is_macos and self.is_default_shell
 
-    @property
     def final_env(self) -> Dict[str, str]:
         from kitty.options.utils import DELETE_ENV_VAR
-        env: Optional[Dict[str, str]] = getattr(self, '_final_env', None)
-        if env is None:
-            env = self._final_env = default_env().copy()
-            if is_macos and env.get('LC_CTYPE') == 'UTF-8' and not sys._xoptions.get(
-                    'lc_ctype_before_python') and not getattr(default_env, 'lc_ctype_set_by_user', False):
-                del env['LC_CTYPE']
-            env.update(self.env)
-            env['TERM'] = fast_data_types.get_options().term
-            env['COLORTERM'] = 'truecolor'
-            env['KITTY_PID'] = getpid()
-            if self.cwd:
-                # needed in case cwd is a symlink, in which case shells
-                # can use it to display the current directory name rather
-                # than the resolved path
-                env['PWD'] = self.cwd
-            tdir = checked_terminfo_dir()
-            if tdir:
-                env['TERMINFO'] = tdir
-            env['KITTY_INSTALLATION_DIR'] = kitty_base_dir
-            opts = fast_data_types.get_options()
-            if 'disabled' not in opts.shell_integration:
-                from .shell_integration import modify_shell_environ
-                modify_shell_environ(opts, env, self.argv)
-            env = {k: v for k, v in env.items() if v is not DELETE_ENV_VAR}
+        env = default_env().copy()
+        opts = fast_data_types.get_options()
+        boss = fast_data_types.get_boss()
+        if is_macos and env.get('LC_CTYPE') == 'UTF-8' and not getattr(sys, 'kitty_run_data').get(
+                'lc_ctype_before_python') and not getattr(default_env, 'lc_ctype_set_by_user', False):
+            del env['LC_CTYPE']
+        env.update(self.env)
+        env['TERM'] = opts.term
+        env['COLORTERM'] = 'truecolor'
+        env['KITTY_PID'] = getpid()
+        env['KITTY_PUBLIC_KEY'] = boss.encryption_public_key
+        if self.add_listen_on_env_var and boss.listening_on:
+            env['KITTY_LISTEN_ON'] = boss.listening_on
+        else:
+            env.pop('KITTY_LISTEN_ON', None)
+        if self.cwd:
+            # needed in case cwd is a symlink, in which case shells
+            # can use it to display the current directory name rather
+            # than the resolved path
+            env['PWD'] = self.cwd
+        tdir = checked_terminfo_dir()
+        if tdir:
+            env['TERMINFO'] = tdir
+        env['KITTY_INSTALLATION_DIR'] = kitty_base_dir
+        if opts.forward_stdio:
+            env['KITTY_STDIO_FORWARDED'] = '3'
+        self.unmodified_argv = list(self.argv)
+        if not self.should_run_via_run_shell_kitten and 'disabled' not in opts.shell_integration:
+            from .shell_integration import modify_shell_environ
+            modify_shell_environ(opts, env, self.argv)
+        env = {k: v for k, v in env.items() if v is not DELETE_ENV_VAR}
+        if self.is_clone_launch:
+            env['KITTY_IS_CLONE_LAUNCH'] = self.is_clone_launch
+            self.is_clone_launch = '1'  # free memory
+        else:
+            env.pop('KITTY_IS_CLONE_LAUNCH', None)
         return env
 
     def fork(self) -> Optional[int]:
         if self.forked:
             return None
+        opts = fast_data_types.get_options()
         self.forked = True
         master, slave = openpty()
         stdin, self.stdin = self.stdin, None
         ready_read_fd, ready_write_fd = os.pipe()
-        remove_cloexec(ready_read_fd)
+        os.set_inheritable(ready_write_fd, False)
+        os.set_inheritable(ready_read_fd, True)
         if stdin is not None:
             stdin_read_fd, stdin_write_fd = os.pipe()
-            remove_cloexec(stdin_read_fd)
+            os.set_inheritable(stdin_write_fd, False)
+            os.set_inheritable(stdin_read_fd, True)
         else:
             stdin_read_fd = stdin_write_fd = -1
-        env = tuple(f'{k}={v}' for k, v in self.final_env.items())
+        final_env = self.final_env()
+        env = tuple(f'{k}={v}' for k, v in final_env.items())
         argv = list(self.argv)
-        exe = argv[0]
-        if is_macos and exe == shell_path:
+        if self.should_run_via_run_shell_kitten:
             # bash will only source ~/.bash_profile if it detects it is a login
             # shell (see the invocation section of the bash man page), which it
             # does if argv[0] is prefixed by a hyphen see
@@ -270,9 +291,23 @@ class Child:
             # https://github.com/kovidgoyal/kitty/issues/1870
             # xterm, urxvt, konsole and gnome-terminal do not do it in my
             # testing.
-            argv[0] = (f'-{exe.split("/")[-1]}')
-        exe = which(exe) or exe
-        pid = fast_data_types.spawn(exe, self.cwd, tuple(argv), env, master, slave, stdin_read_fd, stdin_write_fd, ready_read_fd, ready_write_fd)
+            import shlex
+            ksi = ' '.join(opts.shell_integration)
+            if ksi == 'invalid':
+                ksi = 'enabled'
+            argv = [kitten_exe(), 'run-shell', '--shell', shlex.join(argv), '--shell-integration', ksi]
+            if is_macos:
+                # In addition for getlogin() to work we need to run the shell
+                # via the /usr/bin/login wrapper, sigh.
+                # https://github.com/kovidgoyal/kitty/issues/6511
+                import pwd
+                user = pwd.getpwuid(os.geteuid()).pw_name
+                argv = ['/usr/bin/login', '-f', '-l', '-p', user] + argv
+        self.final_exe = which(argv[0]) or argv[0]
+        self.final_argv0 = argv[0]
+        pid = fast_data_types.spawn(
+            self.final_exe, self.cwd, tuple(argv), env, master, slave, stdin_read_fd, stdin_write_fd,
+            ready_read_fd, ready_write_fd, tuple(handled_signals), kitten_exe(), opts.forward_stdio)
         os.close(slave)
         self.pid = pid
         self.child_fd = master
@@ -282,13 +317,28 @@ class Child:
         os.close(ready_read_fd)
         self.terminal_ready_fd = ready_write_fd
         if self.child_fd is not None:
-            remove_blocking(self.child_fd)
+            os.set_blocking(self.child_fd, False)
         return pid
 
+    def __del__(self) -> None:
+        fd = getattr(self, 'terminal_ready_fd', -1)
+        if fd > -1:
+            os.close(fd)
+        self.terminal_ready_fd = -1
+
     def mark_terminal_ready(self) -> None:
         os.close(self.terminal_ready_fd)
         self.terminal_ready_fd = -1
 
+    def cmdline_of_pid(self, pid: int) -> List[str]:
+        try:
+            ans = cmdline_of_pid(pid)
+        except Exception:
+            ans = []
+        if pid == self.pid and (not ans):
+            ans = list(self.argv)
+        return ans
+
     @property
     def foreground_processes(self) -> List[ProcessDesc]:
         if self.child_fd is None:
@@ -300,7 +350,7 @@ class Child:
             def process_desc(pid: int) -> ProcessDesc:
                 ans: ProcessDesc = {'pid': pid, 'cmdline': None, 'cwd': None}
                 with suppress(Exception):
-                    ans['cmdline'] = cmdline_of_process(pid)
+                    ans['cmdline'] = self.cmdline_of_pid(pid)
                 with suppress(Exception):
                     ans['cwd'] = cwd_of_process(pid) or None
                 return ans
@@ -313,7 +363,7 @@ class Child:
     def cmdline(self) -> List[str]:
         try:
             assert self.pid is not None
-            return cmdline_of_process(self.pid) or list(self.argv)
+            return self.cmdline_of_pid(self.pid) or list(self.argv)
         except Exception:
             return list(self.argv)
 
@@ -321,7 +371,7 @@ class Child:
     def foreground_cmdline(self) -> List[str]:
         try:
             assert self.pid_for_cwd is not None
-            return cmdline_of_process(self.pid_for_cwd) or self.cmdline
+            return self.cmdline_of_pid(self.pid_for_cwd) or self.cmdline
         except Exception:
             return self.cmdline
 
@@ -340,8 +390,7 @@ class Child:
             return cwd_of_process(self.pid)
         return None
 
-    @property
-    def pid_for_cwd(self) -> Optional[int]:
+    def get_pid_for_cwd(self, oldest: bool = False) -> Optional[int]:
         with suppress(Exception):
             assert self.child_fd is not None
             pgrp = os.tcgetpgrp(self.child_fd)
@@ -357,25 +406,62 @@ class Child:
                 # vim
                 # With this script , the foreground process group will contain
                 # both the bash instance running the script and vim.
-                return max(foreground_processes)
+                return min(foreground_processes) if oldest else max(foreground_processes)
         return self.pid
 
     @property
-    def foreground_cwd(self) -> Optional[str]:
+    def pid_for_cwd(self) -> Optional[int]:
+        return self.get_pid_for_cwd()
+
+    def get_foreground_cwd(self, oldest: bool = False) -> Optional[str]:
         with suppress(Exception):
-            assert self.pid_for_cwd is not None
-            return cwd_of_process(self.pid_for_cwd) or None
+            pid = self.get_pid_for_cwd(oldest)
+            if pid is not None:
+                return cwd_of_process(pid) or None
+        return None
+
+    def get_foreground_exe(self, oldest: bool = False) -> Optional[str]:
+        with suppress(Exception):
+            pid = self.get_pid_for_cwd(oldest)
+            if pid is not None:
+                c = cmdline_of_pid(pid)
+                if c:
+                    return c[0]
         return None
 
+    @property
+    def foreground_cwd(self) -> Optional[str]:
+        return self.get_foreground_cwd()
+
     @property
     def foreground_environ(self) -> Dict[str, str]:
-        try:
-            assert self.pid_for_cwd is not None
-            return environ_of_process(self.pid_for_cwd)
-        except Exception:
-            try:
-                assert self.pid is not None
-                return environ_of_process(self.pid)
-            except Exception:
-                pass
+        pid = self.pid_for_cwd
+        if pid is not None:
+            with suppress(Exception):
+                return environ_of_process(pid)
+        pid = self.pid
+        if pid is not None:
+            with suppress(Exception):
+                return environ_of_process(pid)
         return {}
+
+    def send_signal_for_key(self, key_num: bytes) -> bool:
+        import signal
+        import termios
+        if self.child_fd is None:
+            return False
+        t = termios.tcgetattr(self.child_fd)
+        if not t[3] & termios.ISIG:
+            return False
+        cc = t[-1]
+        if key_num == cc[termios.VINTR]:
+            s = signal.SIGINT
+        elif key_num == cc[termios.VSUSP]:
+            s = signal.SIGTSTP
+        elif key_num == cc[termios.VQUIT]:
+            s = signal.SIGQUIT
+        else:
+            return False
+        pgrp = os.tcgetpgrp(self.child_fd)
+        os.killpg(pgrp, s)
+        return True
