@@ -50,18 +50,20 @@ from .options.types import Options
 from .options.utils import DELETE_ENV_VAR
 from .os_window_size import initial_window_size_func
 from .session import create_sessions, get_os_window_sizing_data
+from .shaders import CompileError, load_shader_programs
 from .types import SingleInstanceData
 from .utils import (
     cleanup_ssh_control_masters,
     detach,
     expandvars,
+    get_custom_window_icon,
     log_error,
     parse_os_window_state,
+    safe_mtime,
     single_instance,
     startup_notification_handler,
     unix_socket_paths,
 )
-from .window import load_shader_programs
 
 
 def set_custom_ibeam_cursor() -> None:
@@ -125,8 +127,11 @@ def talk_to_instance(args: CLIOptions) -> None:
 
 
 def load_all_shaders(semi_transparent: bool = False) -> None:
-    load_shader_programs(semi_transparent)
-    load_borders_program()
+    try:
+        load_shader_programs(semi_transparent)
+        load_borders_program()
+    except CompileError as err:
+        raise SystemExit(err)
 
 
 def init_glfw_module(glfw_module: str, debug_keyboard: bool = False, debug_rendering: bool = False) -> None:
@@ -145,7 +150,7 @@ def get_macos_shortcut_for(
 ) -> Optional[SingleKey]:
     # for maximum robustness we should use opts.alias_map to resolve
     # aliases however this requires parsing everything on startup which could
-    # be potentially slow. Lets just hope the user doesnt alias these
+    # be potentially slow. Lets just hope the user doesn't alias these
     # functions.
     ans = None
     candidates = []
@@ -167,43 +172,49 @@ def get_macos_shortcut_for(
     return ans
 
 
-def safe_mtime(path: str) -> Optional[float]:
-    with suppress(OSError):
-        return os.path.getmtime(path)
-    return None
-
-
 def set_macos_app_custom_icon() -> None:
-    for name in ('kitty.app.icns', 'kitty.app.png'):
-        icon_path = os.path.join(config_dir, name)
-        custom_icon_mtime = safe_mtime(icon_path)
-        if custom_icon_mtime is not None:
-            from .fast_data_types import cocoa_set_app_icon, cocoa_set_dock_icon
-            krd = getattr(sys, 'kitty_run_data')
-            bundle_path = os.path.dirname(os.path.dirname(krd.get('bundle_exe_dir')))
-            icon_sentinel = os.path.join(bundle_path, 'Icon\r')
-            sentinel_mtime = safe_mtime(icon_sentinel)
-            if sentinel_mtime is None or sentinel_mtime < custom_icon_mtime:
-                try:
-                    cocoa_set_app_icon(icon_path, bundle_path)
-                except (FileNotFoundError, OSError) as e:
-                    log_error(str(e))
-                    log_error('Failed to set custom app icon, ignoring')
-            # macOS Dock does not reload icons until it is restarted, so we set
-            # the application icon here. This will revert when kitty quits, but
-            # cant be helped since there appears to be no way to get the dock
-            # to reload short of killing it.
-            cocoa_set_dock_icon(icon_path)
-            break
+    custom_icon_mtime, custom_icon_path = get_custom_window_icon()
+    if custom_icon_mtime is not None and custom_icon_path is not None:
+        from .fast_data_types import cocoa_set_app_icon, cocoa_set_dock_icon
+        krd = getattr(sys, 'kitty_run_data')
+        bundle_path = os.path.dirname(os.path.dirname(krd.get('bundle_exe_dir')))
+        icon_sentinel = os.path.join(bundle_path, 'Icon\r')
+        sentinel_mtime = safe_mtime(icon_sentinel)
+        if sentinel_mtime is None or sentinel_mtime < custom_icon_mtime:
+            try:
+                cocoa_set_app_icon(custom_icon_path, bundle_path)
+            except (FileNotFoundError, OSError) as e:
+                log_error(str(e))
+                log_error('Failed to set custom app icon, ignoring')
+        # macOS Dock does not reload icons until it is restarted, so we set
+        # the application icon here. This will revert when kitty quits, but
+        # can't be helped since there appears to be no way to get the dock
+        # to reload short of killing it.
+        cocoa_set_dock_icon(custom_icon_path)
+
+
+def get_icon128_path(base_path: str) -> str:
+    # max icon size on X11 64bits is 128x128
+    path, ext = os.path.splitext(base_path)
+    return f'{path}-128{ext}'
 
 
 def set_x11_window_icon() -> None:
-    # max icon size on X11 64bits is 128x128
-    path, ext = os.path.splitext(logo_png_file)
-    set_default_window_icon(f'{path}-128{ext}')
+    custom_icon_path = get_custom_window_icon()[1]
+    try:
+        if custom_icon_path is not None:
+            custom_icon128_path = get_icon128_path(custom_icon_path)
+            if safe_mtime(custom_icon128_path) is None:
+                set_default_window_icon(custom_icon_path)
+            else:
+                set_default_window_icon(custom_icon128_path)
+        else:
+            set_default_window_icon(get_icon128_path(logo_png_file))
+    except ValueError as err:
+        log_error(err)
 
 
-def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
+def set_cocoa_global_shortcuts(opts: Options) -> Dict[str, SingleKey]:
     global_shortcuts: Dict[str, SingleKey] = {}
     if is_macos:
         from collections import defaultdict
@@ -230,13 +241,20 @@ def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ())
         val = get_macos_shortcut_for(func_map, f'open_url {website_url()}', lookup_name='open_kitty_website')
         if val is not None:
             global_shortcuts['open_kitty_website'] = val
+    return global_shortcuts
 
+
+def _run_app(opts: Options, args: CLIOptions, bad_lines: Sequence[BadLine] = ()) -> None:
+    if is_macos:
+        global_shortcuts = set_cocoa_global_shortcuts(opts)
         if opts.macos_custom_beam_cursor:
             set_custom_ibeam_cursor()
         set_macos_app_custom_icon()
+    else:
+        global_shortcuts = {}
+        if not is_wayland():  # no window icons on wayland
+            set_x11_window_icon()
 
-    if not is_wayland() and not is_macos:  # no window icons on wayland
-        set_x11_window_icon()
     with cached_values_for(run_app.cached_values_name) as cached_values:
         startup_sessions = tuple(create_sessions(opts, args, default_session=opts.startup_session))
         wincls = (startup_sessions[0].os_window_class if startup_sessions else '') or args.cls or appname
@@ -291,15 +309,20 @@ def ensure_macos_locale() -> None:
     # https://github.com/kovidgoyal/kitty/issues/90
     from .fast_data_types import cocoa_get_lang, locale_is_valid
     if 'LANG' not in os.environ:
-        lang = cocoa_get_lang()
-        if lang is not None:
-            if not locale_is_valid(lang):
-                if lang.startswith('en_'):
-                    lang = 'en_US'
-                else:
-                    log_error(f'Could not set LANG Cocoa returns language as: {lang}')
-            os.environ['LANG'] = f'{lang}.UTF-8'
-            set_LANG_in_default_env(os.environ['LANG'])
+        lang_code, country_code, identifier = cocoa_get_lang()
+        lang = 'en_US'
+        if identifier and locale_is_valid(identifier):
+            lang = identifier
+        elif lang_code and country_code and locale_is_valid(f'{lang_code}_{country_code}'):
+            lang = f'{lang_code}_{country_code}'
+        elif lang_code:
+            if lang_code != 'en':
+                with suppress(OSError):
+                    found = sorted(x for x in os.listdir('/usr/share/locale') if x.startswith(f'{lang_code}_'))
+                    if found:
+                        lang = found[0].partition('.')[0]
+        os.environ['LANG'] = f'{lang}.UTF-8'
+        set_LANG_in_default_env(os.environ['LANG'])
 
 
 @contextmanager
@@ -342,8 +365,10 @@ def macos_cmdline(argv_args: List[str]) -> List[str]:
 
 
 def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
+    if from_config_file and listen_on == 'none':
+        return ''
     listen_on = expandvars(listen_on)
-    if '{kitty_pid}' not in listen_on and from_config_file:
+    if '{kitty_pid}' not in listen_on and from_config_file and listen_on.startswith('unix:'):
         listen_on += '-{kitty_pid}'
     listen_on = listen_on.replace('{kitty_pid}', str(os.getpid()))
     if listen_on.startswith('unix:'):
@@ -354,6 +379,9 @@ def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
             elif not os.path.isabs(path):
                 import tempfile
                 listen_on = f'unix:{os.path.join(tempfile.gettempdir(), path)}'
+    elif listen_on.startswith('tcp:') or listen_on.startswith('tcp6:'):
+        if from_config_file:  # use a random port
+            listen_on = ':'.join(listen_on.split(':', 2)[:2]) + ':0'
     return listen_on
 
 
@@ -416,7 +444,7 @@ def setup_manpath(env: Dict[str, str]) -> None:
 
 def setup_environment(opts: Options, cli_opts: CLIOptions) -> None:
     from_config_file = False
-    if not cli_opts.listen_on and opts.listen_on.startswith('unix:'):
+    if not cli_opts.listen_on:
         cli_opts.listen_on = opts.listen_on
         from_config_file = True
     if cli_opts.listen_on:
