@@ -594,6 +594,8 @@ def optional_edge_width(x: str) -> FloatEdges:
 def hide_window_decorations(x: str) -> int:
     if x == 'titlebar-only':
         return 0b10
+    if x == 'titlebar-and-corners':
+        return 0b100
     if to_bool(x):
         return 0b01
     return 0b00
@@ -604,9 +606,16 @@ def resize_draw_strategy(x: str) -> int:
     return cmap.get(x.lower(), 0)
 
 
+def resize_debounce_time(x: str) -> Tuple[float, float]:
+    parts = x.split(maxsplit=1)
+    if len(parts) == 1:
+        return positive_float(parts[0]), 0.5
+    return positive_float(parts[0]), positive_float(parts[1])
+
+
 def visual_window_select_characters(x: str) -> str:
     import string
-    valid_characters = string.digits + string.ascii_uppercase
+    valid_characters = string.digits + string.ascii_uppercase + "-=[]\\;',./`"
     ans = x.upper()
     ans_chars = set(ans)
     if not ans_chars.issubset(set(valid_characters)):
@@ -854,7 +863,7 @@ def store_multiple(val: str, current_val: Container[str]) -> Iterable[Tuple[str,
         yield val, val
 
 
-allowed_shell_integration_values = frozenset({'enabled', 'disabled', 'no-rc', 'no-cursor', 'no-title', 'no-prompt-mark', 'no-complete', 'no-cwd'})
+allowed_shell_integration_values = frozenset({'enabled', 'disabled', 'no-rc', 'no-cursor', 'no-title', 'no-prompt-mark', 'no-complete', 'no-cwd', 'no-sudo'})
 
 
 def shell_integration(x: str) -> FrozenSet[str]:
