@@ -2,6 +2,7 @@
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
+import re
 import stat
 import weakref
 from collections import deque
@@ -9,27 +10,55 @@ from contextlib import suppress
 from operator import attrgetter
 from time import monotonic
 from typing import (
-    Any, Deque, Dict, Generator, Iterable, Iterator, List, NamedTuple,
-    Optional, Pattern, Sequence, Set, Tuple, Union, cast
+    Any,
+    Callable,
+    Deque,
+    Dict,
+    Generator,
+    Iterable,
+    Iterator,
+    List,
+    NamedTuple,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Union,
 )
 
 from .borders import Border, Borders
 from .child import Child
 from .cli_stub import CLIOptions
-from .constants import appname, kitty_exe
+from .constants import appname
 from .fast_data_types import (
-    GLFW_MOUSE_BUTTON_LEFT, GLFW_MOUSE_BUTTON_MIDDLE, GLFW_PRESS, GLFW_RELEASE,
-    add_tab, attach_window, detach_window, get_boss, get_click_interval,
-    get_options, mark_tab_bar_dirty, next_window_id, remove_tab, remove_window,
-    ring_bell, set_active_tab, set_active_window, swap_tabs,
-    sync_os_window_title
+    GLFW_MOUSE_BUTTON_LEFT,
+    GLFW_MOUSE_BUTTON_MIDDLE,
+    GLFW_PRESS,
+    GLFW_RELEASE,
+    add_tab,
+    attach_window,
+    current_focused_os_window_id,
+    detach_window,
+    get_boss,
+    get_click_interval,
+    get_options,
+    last_focused_os_window_id,
+    mark_tab_bar_dirty,
+    next_window_id,
+    remove_tab,
+    remove_window,
+    ring_bell,
+    set_active_tab,
+    set_active_window,
+    swap_tabs,
+    sync_os_window_title,
 )
 from .layout.base import Layout
 from .layout.interface import create_layout_object_for, evict_cached_layouts
 from .tab_bar import TabBar, TabBarData
 from .types import ac
 from .typing import EdgeLiteral, SessionTab, SessionType, TypedDict
-from .utils import log_error, platform_window_id, resolved_shell
+from .utils import cmdline_for_hold, log_error, platform_window_id, resolved_shell
 from .window import CwdRequest, Watchers, Window, WindowDict
 from .window_list import WindowList
 
@@ -45,12 +74,14 @@ class TabMouseEvent(NamedTuple):
 class TabDict(TypedDict):
     id: int
     is_focused: bool
+    is_active: bool
     title: str
     layout: str
     layout_state: Dict[str, Any]
     layout_opts: Dict[str, Any]
     enabled_layouts: List[str]
     windows: List[WindowDict]
+    groups: List[Dict[str, Any]]
     active_window_history: List[int]
 
 
@@ -94,6 +125,7 @@ class Tab:  # {{{
     active_bg: Optional[int] = None
     inactive_fg: Optional[int] = None
     inactive_bg: Optional[int] = None
+    confirm_close_window_id: int = 0
 
     def __init__(
         self,
@@ -112,7 +144,7 @@ class Tab:  # {{{
         self.name = getattr(session_tab, 'name', '')
         self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or get_options().enabled_layouts]
         self.borders = Borders(self.os_window_id, self.id)
-        self.windows = WindowList(self)
+        self.windows: WindowList = WindowList(self)
         self._last_used_layout: Optional[str] = None
         self._current_layout_name: Optional[str] = None
         self.cwd = self.args.directory
@@ -138,9 +170,10 @@ class Tab:  # {{{
             self._set_current_layout(self.enabled_layouts[0])
         self.relayout()
 
-    def apply_options(self) -> None:
+    def apply_options(self, is_active: bool) -> None:
+        aw = self.active_window
         for window in self:
-            window.apply_options()
+            window.apply_options(is_active and aw is window)
         self.set_enabled_layouts(get_options().enabled_layouts)
 
     def take_over_from(self, other_tab: 'Tab') -> None:
@@ -167,12 +200,16 @@ class Tab:  # {{{
         self.mark_tab_bar_dirty()
 
     def startup(self, session_tab: 'SessionTab') -> None:
-        for cmd in session_tab.windows:
-            if isinstance(cmd, SpecialWindowInstance):
-                self.new_special_window(cmd)
+        for window in session_tab.windows:
+            spec = window.launch_spec
+            if isinstance(spec, SpecialWindowInstance):
+                self.new_special_window(spec)
             else:
                 from .launch import launch
-                launch(get_boss(), cmd.opts, cmd.args, target_tab=self, force_target_tab=True)
+                launch(get_boss(), spec.opts, spec.args, target_tab=self, force_target_tab=True)
+            if window.resize_spec is not None:
+                self.resize_window(*window.resize_spec)
+
         self.windows.set_active_window_group_for(self.windows.all_windows[session_tab.active_window_idx])
 
     def serialize_state(self) -> Dict[str, Any]:
@@ -206,11 +243,16 @@ class Tab:  # {{{
 
     @property
     def active_window_for_cwd(self) -> Optional[Window]:
-        return self.windows.active_group_base
+        return self.windows.active_group_main
 
     @property
     def title(self) -> str:
-        return cast(str, getattr(self.active_window, 'title', appname))
+        w = self.active_window
+        return w.title if w else appname
+
+    @property
+    def effective_title(self) -> str:
+        return self.name or self.title
 
     @property
     def number_of_windows_with_running_programs(self) -> int:
@@ -220,6 +262,14 @@ class Tab:  # {{{
                 ans += 1
         return ans
 
+    def get_cwd_of_active_window(self, oldest: bool = False) -> Optional[str]:
+        w = self.active_window
+        return w.get_cwd_of_child(oldest) if w else None
+
+    def get_exe_of_active_window(self, oldest: bool = False) -> Optional[str]:
+        w = self.active_window
+        return w.get_exe_of_child(oldest) if w else None
+
     def set_title(self, title: str) -> None:
         self.name = title or ''
         self.mark_tab_bar_dirty()
@@ -241,21 +291,18 @@ class Tab:  # {{{
     def relayout_borders(self) -> None:
         tm = self.tab_manager_ref()
         if tm is not None:
-            w = self.active_window
             ly = self.current_layout
             self.borders(
                 all_windows=self.windows,
                 current_layout=ly, tab_bar_rects=tm.tab_bar_rects,
                 draw_window_borders=(ly.needs_window_borders and self.windows.num_visble_groups > 1) or ly.must_draw_borders
             )
-            if w is not None:
-                w.change_titlebar_color()
 
     def create_layout_object(self, name: str) -> Layout:
         return create_layout_object_for(name, self.os_window_id, self.id)
 
-    @ac('lay', 'Go to the next enabled layout')
-    def next_layout(self) -> None:
+    @ac('lay', 'Go to the next enabled layout. Can optionally supply an integer to jump by the specified number.')
+    def next_layout(self, delta: int = 1) -> None:
         if len(self.enabled_layouts) > 1:
             for i, layout_name in enumerate(self.enabled_layouts):
                 if layout_name == self.current_layout.full_name:
@@ -263,7 +310,10 @@ class Tab:  # {{{
                     break
             else:
                 idx = -1
-            nl = self.enabled_layouts[(idx + 1) % len(self.enabled_layouts)]
+            if abs(delta) >= len(self.enabled_layouts):
+                mult = -1 if delta < 0 else 1
+                delta = mult * (abs(delta) % len(self.enabled_layouts))
+            nl = self.enabled_layouts[(idx + delta + len(self.enabled_layouts)) % len(self.enabled_layouts)]
             self._set_current_layout(nl)
             self.relayout()
 
@@ -275,20 +325,47 @@ class Tab:  # {{{
 
     @ac('lay', '''
         Switch to the named layout
+        In case there are multiple layouts with the same name and different options,
+        specify the full layout definition or a unique prefix of the full definition.
 
         For example::
 
             map f1 goto_layout tall
+            map f2 goto_layout fat:bias=20
         ''')
     def goto_layout(self, layout_name: str, raise_exception: bool = False) -> None:
         layout_name = layout_name.lower()
-        if layout_name not in self.enabled_layouts:
-            if raise_exception:
-                raise ValueError(layout_name)
-            log_error(f'Unknown or disabled layout: {layout_name}')
-            return
-        self._set_current_layout(layout_name)
-        self.relayout()
+        q, has_colon, rest = layout_name.partition(':')
+        matches = []
+        prefix_matches = []
+        matched_layout = ''
+        for candidate in self.enabled_layouts:
+            x, _, _ = candidate.partition(':')
+            if x == q:
+                if candidate == layout_name:
+                    matched_layout = candidate
+                    break
+                if candidate.startswith(layout_name):
+                    prefix_matches.append(candidate)
+                matches.append(x)
+
+        if not matched_layout:
+            if len(prefix_matches) == 1:
+                matched_layout = prefix_matches[0]
+            elif len(matches) == 1:
+                matched_layout = matches[0]
+        if matched_layout:
+            self._set_current_layout(matched_layout)
+            self.relayout()
+        else:
+            if len(matches) == 0:
+                if raise_exception:
+                    raise ValueError(layout_name)
+                log_error(f'Unknown or disabled layout: {layout_name}')
+            elif len(matches) != 1:
+                if raise_exception:
+                    raise ValueError(layout_name)
+                log_error(f'Multiple layouts match: {layout_name}')
 
     @ac('lay', '''
         Toggle the named layout
@@ -340,7 +417,8 @@ class Tab:  # {{{
     def layout_action(self, action_name: str, args: Sequence[str]) -> None:
         ret = self.current_layout.layout_action(action_name, args, self.windows)
         if ret is None:
-            ring_bell()
+            if get_options().enable_audio_bell:
+                ring_bell()
             return
         self.relayout()
 
@@ -352,7 +430,8 @@ class Tab:  # {{{
         cwd_from: Optional[CwdRequest] = None,
         cwd: Optional[str] = None,
         env: Optional[Dict[str, str]] = None,
-        allow_remote_control: bool = False
+        is_clone_launch: str = '',
+        add_listen_on_env_var: bool = True,
     ) -> Child:
         check_for_suitability = True
         if cmd is None:
@@ -386,6 +465,8 @@ class Tab:  # {{{
                         cmd = resolved_shell(get_options())
                     elif not is_executable:
                         import shlex
+
+                        from .utils import which
                         with suppress(OSError):
                             with open(old_exe) as f:
                                 if f.read(2) == '#!':
@@ -393,7 +474,8 @@ class Tab:  # {{{
                                     cmd[:0] = shlex.split(line)
                                 else:
                                     cmd[:0] = [resolved_shell(get_options())[0]]
-                                cmd[:0] = [kitty_exe(), '+hold']
+                                cmd[0] = which(cmd[0]) or cmd[0]
+                                cmd = cmdline_for_hold(cmd)
         fenv: Dict[str, str] = {}
         if env:
             fenv.update(env)
@@ -401,7 +483,7 @@ class Tab:  # {{{
         pwid = platform_window_id(self.os_window_id)
         if pwid is not None:
             fenv['WINDOWID'] = str(pwid)
-        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
+        ans = Child(cmd, cwd or self.cwd, stdin, fenv, cwd_from, is_clone_launch=is_clone_launch, add_listen_on_env_var=add_listen_on_env_var)
         ans.fork()
         return ans
 
@@ -425,13 +507,18 @@ class Tab:  # {{{
         allow_remote_control: bool = False,
         marker: Optional[str] = None,
         watchers: Optional[Watchers] = None,
-        overlay_behind: bool = False
+        overlay_behind: bool = False,
+        is_clone_launch: str = '',
+        remote_control_passwords: Optional[Dict[str, Sequence[str]]] = None,
     ) -> Window:
         child = self.launch_child(
-            use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env, allow_remote_control=allow_remote_control)
+            use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env,
+            is_clone_launch=is_clone_launch, add_listen_on_env_var=False if allow_remote_control and remote_control_passwords else True
+        )
         window = Window(
             self, child, self.args, override_title=override_title,
-            copy_colors_from=copy_colors_from, watchers=watchers
+            copy_colors_from=copy_colors_from, watchers=watchers,
+            allow_remote_control=allow_remote_control, remote_control_passwords=remote_control_passwords
         )
         # Must add child before laying out so that resize_pty succeeds
         get_boss().add_child(window)
@@ -470,6 +557,14 @@ class Tab:  # {{{
     def move_window_to_top_of_group(self, window: Window) -> bool:
         return self.windows.move_window_to_top_of_group(window)
 
+    def overlay_parent(self, window: Window) -> Optional[Window]:
+        prev: Optional[Window] = None
+        for x in self.windows.windows_in_group_of(window):
+            if x is window:
+                break
+            prev = x
+        return prev
+
     def remove_window(self, window: Window, destroy: bool = True) -> None:
         self.windows.remove_window(window)
         if destroy:
@@ -494,8 +589,8 @@ class Tab:  # {{{
         attach_window(self.os_window_id, self.id, window.id)
         self._add_window(window)
 
-    def set_active_window(self, x: Union[Window, int]) -> None:
-        self.windows.set_active_window_group_for(x)
+    def set_active_window(self, x: Union[Window, int], for_keep_focus: Optional[Window] = None) -> None:
+        self.windows.set_active_window_group_for(x, for_keep_focus=for_keep_focus)
 
     def get_nth_window(self, n: int) -> Optional[Window]:
         if self.windows:
@@ -678,15 +773,53 @@ class Tab:  # {{{
     def move_window_backward(self) -> None:
         self.move_window(-1)
 
-    def list_windows(self, active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[WindowDict, None, None]:
+    def list_windows(self, self_window: Optional[Window] = None, window_filter: Optional[Callable[[Window], bool]] = None) -> Generator[WindowDict, None, None]:
+        active_window = self.active_window
         for w in self:
-            yield w.as_dict(is_focused=w is active_window, is_self=w is self_window)
+            if window_filter is None or window_filter(w):
+                yield w.as_dict(
+                    is_active=w is active_window,
+                    is_focused=w.os_window_id == current_focused_os_window_id() and w is active_window,
+                    is_self=w is self_window)
 
-    def matches(self, field: str, pat: 'Pattern[str]') -> bool:
-        if field == 'id':
-            return bool(pat.pattern == str(self.id))
+    def list_groups(self) -> List[Dict[str, Any]]:
+        return [g.as_simple_dict() for g in self.windows.groups]
+
+    def matches_query(self, field: str, query: str, active_tab_manager: Optional['TabManager'] = None) -> bool:
         if field == 'title':
-            return pat.search(self.name or self.title) is not None
+            return re.search(query, self.effective_title) is not None
+        if field == 'id':
+            return query == str(self.id)
+        if field in ('window_id', 'window_title'):
+            field = field.partition('_')[-1]
+            for w in self:
+                if w.matches_query(field, query):
+                    return True
+            return False
+        if field == 'index':
+            if active_tab_manager and len(active_tab_manager.tabs):
+                idx = (int(query) + len(active_tab_manager.tabs)) % len(active_tab_manager.tabs)
+                return active_tab_manager.tabs[idx] is self
+            return False
+        if field == 'recent':
+            if active_tab_manager and len(active_tab_manager.tabs):
+                return self is active_tab_manager.nth_active_tab(int(query))
+            return False
+        if field == 'state':
+            if query == 'active':
+                tm = self.tab_manager_ref()
+                return tm is not None and self is tm.active_tab
+            if query == 'focused':
+                return active_tab_manager is not None and self is active_tab_manager.active_tab and self.os_window_id == last_focused_os_window_id()
+            if query == 'needs_attention':
+                for w in self:
+                    if w.needs_attention:
+                        return True
+            if query == 'parent_active':
+                return active_tab_manager is not None and self.tab_manager_ref() is active_tab_manager
+            if query == 'parent_focused':
+                return active_tab_manager is not None and self.tab_manager_ref() is active_tab_manager and self.os_window_id == last_focused_os_window_id()
+            return False
         return False
 
     def __iter__(self) -> Iterator[Window]:
@@ -709,7 +842,7 @@ class Tab:  # {{{
         self.windows = WindowList(self)
 
     def __repr__(self) -> str:
-        return f'Tab(title={self.name or self.title}, id={hex(id(self))})'
+        return f'Tab(title={self.effective_title}, id={self.id})'
 
     def make_active(self) -> None:
         tm = self.tab_manager_ref()
@@ -720,6 +853,8 @@ class Tab:  # {{{
 
 class TabManager:  # {{{
 
+    confirm_close_window_id: int = 0
+
     def __init__(self, os_window_id: int, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
         self.os_window_id = os_window_id
         self.wm_class = wm_class
@@ -790,8 +925,11 @@ class TabManager:  # {{{
         if visible_before and not self.tab_bar_should_be_visible:
             self.tabbar_visibility_changed()
 
-    def _set_active_tab(self, idx: int) -> None:
-        self.active_tab_idx = idx
+    def _set_active_tab(self, idx: int, store_in_history: bool = True) -> None:
+        if store_in_history:
+            self.active_tab_idx = idx
+        else:
+            self._active_tab_idx = idx
         set_active_tab(self.os_window_id, idx)
 
     def tabbar_visibility_changed(self) -> None:
@@ -826,12 +964,16 @@ class TabManager:  # {{{
             tab.relayout_borders()
         self.mark_tab_bar_dirty()
 
-    def set_active_tab(self, tab: Tab) -> bool:
+    def set_active_tab(self, tab: Tab, for_keep_focus: Optional[Tab] = None) -> bool:
         try:
             idx = self.tabs.index(tab)
         except Exception:
             return False
         self.set_active_tab_idx(idx)
+        h = self.active_tab_history
+        if for_keep_focus and len(h) > 2 and h[-2] == for_keep_focus.id and h[-1] != for_keep_focus.id:
+            h.pop()
+            h.pop()
         return True
 
     def next_tab(self, delta: int = 1) -> None:
@@ -878,19 +1020,29 @@ class TabManager:  # {{{
     def __len__(self) -> int:
         return len(self.tabs)
 
-    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[TabDict, None, None]:
+    def list_tabs(
+        self, self_window: Optional[Window] = None,
+        tab_filter: Optional[Callable[[Tab], bool]] = None,
+        window_filter: Optional[Callable[[Window], bool]] = None
+    ) -> Generator[TabDict, None, None]:
+        active_tab = self.active_tab
         for tab in self:
-            yield {
-                'id': tab.id,
-                'is_focused': tab is active_tab,
-                'title': tab.name or tab.title,
-                'layout': str(tab.current_layout.name),
-                'layout_state': tab.current_layout.layout_state(),
-                'layout_opts': tab.current_layout.layout_opts.serialized(),
-                'enabled_layouts': tab.enabled_layouts,
-                'windows': list(tab.list_windows(active_window, self_window)),
-                'active_window_history': list(tab.windows.active_window_history),
-            }
+            if tab_filter is None or tab_filter(tab):
+                windows = list(tab.list_windows(self_window, window_filter))
+                if windows:
+                    yield {
+                        'id': tab.id,
+                        'is_focused': tab is active_tab and tab.os_window_id == current_focused_os_window_id(),
+                        'is_active': tab is active_tab,
+                        'title': tab.name or tab.title,
+                        'layout': str(tab.current_layout.name),
+                        'layout_state': tab.current_layout.layout_state(),
+                        'layout_opts': tab.current_layout.layout_opts.serialized(),
+                        'enabled_layouts': tab.enabled_layouts,
+                        'windows': windows,
+                        'groups': tab.list_groups(),
+                        'active_window_history': list(tab.windows.active_window_history),
+                    }
 
     def serialize_state(self) -> Dict[str, Any]:
         return {
@@ -956,7 +1108,6 @@ class TabManager:  # {{{
         idx = len(self.tabs)
         orig_active_tab_idx = self.active_tab_idx
         self._add_tab(Tab(self, no_initial_window=True) if empty_tab else Tab(self, special_window=special_window, cwd_from=cwd_from))
-        self._set_active_tab(idx)
         if as_neighbor:
             location = 'after'
         if location == 'neighbor':
@@ -972,40 +1123,45 @@ class TabManager:  # {{{
                 for i in range(idx, desired_idx, -1):
                     self.tabs[i], self.tabs[i-1] = self.tabs[i-1], self.tabs[i]
                     swap_tabs(self.os_window_id, i, i-1)
-                self._set_active_tab(desired_idx)
                 idx = desired_idx
+        self._set_active_tab(idx)
         self.mark_tab_bar_dirty()
         return self.tabs[idx]
 
     def remove(self, tab: Tab) -> None:
         active_tab_before_removal = self.active_tab
+        active_tab_needs_to_change = active_tab_before_removal is tab
         self._remove_tab(tab)
-        active_tab = self.active_tab
-        active_tab_needs_to_change = (active_tab is None and (active_tab_before_removal is None or active_tab_before_removal is tab)) or active_tab is tab
         while True:
             try:
                 self.active_tab_history.remove(tab.id)
             except ValueError:
                 break
 
+        def idx_for_id(tab_id: int) -> int:
+            for idx, qtab in enumerate(self.tabs):
+                if qtab.id == tab_id:
+                    return idx
+            return -1
+
         if active_tab_needs_to_change:
             next_active_tab = -1
-            if get_options().tab_switch_strategy == 'previous':
+            tss = get_options().tab_switch_strategy
+            if tss == 'previous':
                 while self.active_tab_history and next_active_tab < 0:
                     tab_id = self.active_tab_history.pop()
-                    for idx, qtab in enumerate(self.tabs):
-                        if qtab.id == tab_id:
-                            next_active_tab = idx
-                            break
-            elif get_options().tab_switch_strategy == 'left':
+                    next_active_tab = idx_for_id(tab_id)
+            elif tss == 'left':
                 next_active_tab = max(0, self.active_tab_idx - 1)
-            elif get_options().tab_switch_strategy == 'right':
+            elif tss == 'right':
                 next_active_tab = min(self.active_tab_idx, len(self.tabs) - 1)
+            elif tss == 'last':
+                next_active_tab = len(self.tabs) - 1
 
             if next_active_tab < 0:
                 next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
 
-            self._set_active_tab(next_active_tab)
+            self._set_active_tab(next_active_tab, store_in_history=False)
         elif active_tab_before_removal is not None:
             try:
                 idx = self.tabs.index(active_tab_before_removal)
@@ -1021,7 +1177,7 @@ class TabManager:  # {{{
         at = self.active_tab
         ans = []
         for t in self.tabs:
-            title = (t.name or t.title or appname).strip()
+            title = t.name or t.title or appname
             needs_attention = False
             has_activity_since_last_focus = False
             for w in t:
@@ -1030,7 +1186,7 @@ class TabManager:  # {{{
                 if w.has_activity_since_last_focus:
                     has_activity_since_last_focus = True
             ans.append(TabBarData(
-                title, t is at, needs_attention,
+                title, t is at, needs_attention, t.id,
                 len(t), t.num_window_groups, t.current_layout.name or '',
                 has_activity_since_last_focus, t.active_fg, t.active_bg,
                 t.inactive_fg, t.inactive_bg
@@ -1077,8 +1233,9 @@ class TabManager:  # {{{
         del self.tabs
 
     def apply_options(self) -> None:
+        at = self.active_tab
         for tab in self:
-            tab.apply_options()
+            tab.apply_options(at is tab)
         self.tab_bar_hidden = get_options().tab_bar_style == 'hidden'
         self.tab_bar.apply_options()
         self.update_tab_bar_data()
