@@ -4,7 +4,6 @@
 import json
 import os
 import re
-import sys
 import weakref
 from collections import deque
 from contextlib import contextmanager, suppress
@@ -38,38 +37,20 @@ from .constants import (
     appname,
     clear_handled_signals,
     config_dir,
-    is_macos,
+    kitten_exe,
     wakeup_io_loop,
 )
 from .fast_data_types import (
-    BGIMAGE_PROGRAM,
-    BLIT_PROGRAM,
-    CELL_BG_PROGRAM,
-    CELL_FG_PROGRAM,
-    CELL_PROGRAM,
-    CELL_SPECIAL_PROGRAM,
     CURSOR_BEAM,
     CURSOR_BLOCK,
     CURSOR_UNDERLINE,
     DCS,
-    DECORATION,
-    DECORATION_MASK,
-    DIM,
     GLFW_MOD_CONTROL,
-    GRAPHICS_ALPHA_MASK_PROGRAM,
-    GRAPHICS_PREMULT_PROGRAM,
-    GRAPHICS_PROGRAM,
-    MARK,
-    MARK_MASK,
     NO_CURSOR_SHAPE,
-    NUM_UNDERLINE_STYLES,
     OSC,
-    REVERSE,
     SCROLL_FULL,
     SCROLL_LINE,
     SCROLL_PAGE,
-    STRIKETHROUGH,
-    TINT_PROGRAM,
     Color,
     KeyEvent,
     Screen,
@@ -78,19 +59,16 @@ from .fast_data_types import (
     cell_size_for_window,
     click_mouse_cmd_output,
     click_mouse_url,
-    compile_program,
     current_focused_os_window_id,
     encode_key_for_tty,
     get_boss,
     get_click_interval,
     get_options,
-    init_cell_program,
     last_focused_os_window_id,
     mark_os_window_dirty,
     mouse_selection,
     move_cursor_to_mouse_if_in_prompt,
     pt_to_px,
-    set_titlebar_color,
     set_window_logo,
     set_window_padding,
     set_window_render_data,
@@ -112,8 +90,8 @@ from .types import MouseEvent, OverlayType, WindowGeometry, ac, run_once
 from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
     docs_url,
+    key_val_matcher,
     kitty_ansi_sanitizer_pat,
-    load_shaders,
     log_error,
     open_cmd,
     open_url,
@@ -121,6 +99,7 @@ from .utils import (
     path_from_osc7_url,
     resolve_custom_file,
     resolved_shell,
+    sanitize_control_codes,
     sanitize_for_bracketed_paste,
     sanitize_title,
     sanitize_url_for_dispay_to_user,
@@ -166,7 +145,7 @@ class CwdRequest:
             return window.get_cwd_of_root_child() or ''
         return window.get_cwd_of_child(oldest=self.request_type is CwdRequestType.oldest) or ''
 
-    def modify_argv_for_launch_with_cwd(self, argv: List[str]) -> str:
+    def modify_argv_for_launch_with_cwd(self, argv: List[str], env: Optional[Dict[str, str]]=None) -> str:
         window = self.window
         if not window:
             return ''
@@ -176,10 +155,14 @@ class CwdRequest:
             if ssh_kitten_cmdline:
                 run_shell = argv[0] == resolved_shell(get_options())[0]
                 server_args = [] if run_shell else list(argv)
-                from kittens.ssh.utils import set_cwd_in_cmdline, set_server_args_in_cmdline
+                from kittens.ssh.utils import set_cwd_in_cmdline, set_env_in_cmdline, set_server_args_in_cmdline
                 argv[:] = ssh_kitten_cmdline
+                if argv and argv[0] == 'kitten':
+                    argv[0] = kitten_exe()
                 set_cwd_in_cmdline(reported_cwd, argv)
                 set_server_args_in_cmdline(server_args, argv, allocate_tty=not run_shell)
+                if env is not None:
+                    set_env_in_cmdline(env, argv, clone=False)
                 return ''
             if not window.child_is_remote and (self.request_type is CwdRequestType.last_reported or window.at_prompt):
                 return reported_cwd
@@ -222,6 +205,8 @@ class WindowDict(TypedDict):
     is_self: bool
     lines: int
     columns: int
+    user_vars: Dict[str, str]
+    at_prompt: bool
 
 
 class PipeData(TypedDict):
@@ -363,63 +348,6 @@ def as_text(
     return ans
 
 
-def multi_replace(src: str, **replacements: Any) -> str:
-    r = {k: str(v) for k, v in replacements.items()}
-
-    def sub(m: 're.Match[str]') -> str:
-        return r.get(m.group(1), m.group(1))
-
-    return re.sub(r'\{([A-Z_]+)\}', sub, src)
-
-
-class LoadShaderPrograms:
-
-    def __call__(self, semi_transparent: bool = False) -> None:
-        compile_program(BLIT_PROGRAM, *load_shaders('blit'))
-        v, f = load_shaders('cell')
-
-        for which, p in {
-                'SIMPLE': CELL_PROGRAM,
-                'BACKGROUND': CELL_BG_PROGRAM,
-                'SPECIAL': CELL_SPECIAL_PROGRAM,
-                'FOREGROUND': CELL_FG_PROGRAM,
-        }.items():
-            ff = f.replace('{WHICH_PROGRAM}', which)
-            vv = multi_replace(
-                v,
-                WHICH_PROGRAM=which,
-                REVERSE_SHIFT=REVERSE,
-                STRIKE_SHIFT=STRIKETHROUGH,
-                DIM_SHIFT=DIM,
-                DECORATION_SHIFT=DECORATION,
-                MARK_SHIFT=MARK,
-                MARK_MASK=MARK_MASK,
-                DECORATION_MASK=DECORATION_MASK,
-                STRIKE_SPRITE_INDEX=NUM_UNDERLINE_STYLES + 1,
-            )
-            if semi_transparent:
-                vv = vv.replace('#define NOT_TRANSPARENT', '#define TRANSPARENT')
-                ff = ff.replace('#define NOT_TRANSPARENT', '#define TRANSPARENT')
-            compile_program(p, vv, ff)
-
-        v, f = load_shaders('graphics')
-        for which, p in {
-                'SIMPLE': GRAPHICS_PROGRAM,
-                'PREMULT': GRAPHICS_PREMULT_PROGRAM,
-                'ALPHA_MASK': GRAPHICS_ALPHA_MASK_PROGRAM,
-        }.items():
-            ff = f.replace('ALPHA_TYPE', which)
-            compile_program(p, v, ff)
-
-        v, f = load_shaders('bgimage')
-        compile_program(BGIMAGE_PROGRAM, v, f)
-        v, f = load_shaders('tint')
-        compile_program(TINT_PROGRAM, v, f)
-        init_cell_program()
-
-
-load_shader_programs = LoadShaderPrograms()
-
 
 def setup_colors(screen: Screen, opts: Options) -> None:
     screen.color_profile.update_ansi_color_table(build_ansi_color_table(opts))
@@ -597,6 +525,7 @@ class Window:
         self.default_title = os.path.basename(child.argv[0] or appname)
         self.child_title = self.default_title
         self.title_stack: Deque[str] = deque(maxlen=10)
+        self.user_vars: Dict[str, str] = {}
         self.id: int = add_window(tab.os_window_id, tab.id, self.title)
         self.clipboard_request_manager = ClipboardRequestManager(self.id)
         self.margin = EdgeWidths()
@@ -696,10 +625,9 @@ class Window:
             val = round(val)
         return int(val)
 
-    def apply_options(self) -> None:
+    def apply_options(self, is_active: bool) -> None:
         opts = get_options()
         self.update_effective_padding()
-        self.change_titlebar_color()
         setup_colors(self.screen, opts)
 
     @property
@@ -710,20 +638,22 @@ class Window:
         return f'Window(title={self.title}, id={self.id})'
 
     def as_dict(self, is_focused: bool = False, is_self: bool = False, is_active: bool = False) -> WindowDict:
-        return dict(
-            id=self.id,
-            is_focused=is_focused,
-            is_active=is_active,
-            title=self.title,
-            pid=self.child.pid,
-            cwd=self.child.current_cwd or self.child.cwd,
-            cmdline=self.child.cmdline,
-            env=self.child.environ,
-            foreground_processes=self.child.foreground_processes,
-            is_self=is_self,
-            lines=self.screen.lines,
-            columns=self.screen.columns,
-        )
+        return {
+            'id': self.id,
+            'is_focused': is_focused,
+            'is_active': is_active,
+            'title': self.title,
+            'pid': self.child.pid,
+            'cwd': self.child.current_cwd or self.child.cwd,
+            'cmdline': self.child.cmdline,
+            'env': self.child.environ,
+            'foreground_processes': self.child.foreground_processes,
+            'is_self': is_self,
+            'at_prompt': self.at_prompt,
+            'lines': self.screen.lines,
+            'columns': self.screen.columns,
+            'user_vars': self.user_vars,
+        }
 
     def serialize_state(self) -> Dict[str, Any]:
         ans = {
@@ -745,6 +675,8 @@ class Window:
             ans['window_custom_type'] = self.window_custom_type
         if self.overlay_type is not OverlayType.transient:
             ans['overlay_type'] = self.overlay_type.value
+        if self.user_vars:
+            ans['user_vars'] = self.user_vars
         return ans
 
     @property
@@ -769,15 +701,13 @@ class Window:
     def matches(self, field: str, pat: MatchPatternType) -> bool:
         if not pat:
             return False
-        if field == 'env':
-            assert isinstance(pat, tuple)
-            key_pat, val_pat = pat
-            for key, val in self.child.environ.items():
-                if key_pat.search(key) is not None and (
-                        val_pat is None or val_pat.search(val) is not None):
-                    return True
+
+        if isinstance(pat, tuple):
+            if field == 'env':
+                return key_val_matcher(self.child.environ.items(), *pat)
+            if field == 'var':
+                return key_val_matcher(self.user_vars.items(), *pat)
             return False
-        assert not isinstance(pat, tuple)
 
         if field in ('id', 'window_id'):
             return pat.pattern == str(self.id)
@@ -827,7 +757,22 @@ class Window:
             if query == 'overlay_parent':
                 return self_window is not None and self is self_window.overlay_parent
             return False
-        pat = compile_match_query(query, field != 'env')
+        if field == 'neighbor':
+            t = get_boss().active_tab
+            if t is None:
+                return False
+            gid: Optional[int] = None
+            if query == 'left':
+                gid = t.neighboring_group_id("left")
+            elif query == 'right':
+                gid = t.neighboring_group_id("right")
+            elif query == 'top':
+                gid = t.neighboring_group_id("top")
+            elif query == 'bottom':
+                gid = t.neighboring_group_id("bottom")
+            return gid is not None and t.windows.active_window_in_group_id(gid) is self
+
+        pat = compile_match_query(query, field not in ('env', 'var'))
         return self.matches(field, pat)
 
     def set_visible_in_layout(self, val: bool) -> None:
@@ -922,7 +867,30 @@ class Window:
         self.override_title = title or None
         self.title_updated()
 
+    def set_user_var(self, key: str, val: Optional[Union[str, bytes]]) -> None:
+        key = sanitize_control_codes(key).replace('\n', ' ')
+        self.user_vars.pop(key, None)  # ensure key will be newest in user_vars even if already present
+        if len(self.user_vars) > 64:  # dont store too many user vars
+            oldest_key = next(iter(self.user_vars))
+            self.user_vars.pop(oldest_key)
+        if val is not None:
+            if isinstance(val, bytes):
+                val = val.decode('utf-8', 'replace')
+            self.user_vars[key] = sanitize_control_codes(val).replace('\n', ' ')
+
+    # screen callbacks {{{
+
+    def osc_1337(self, raw_data: str) -> None:
+        for record in raw_data.split(';'):
+            key, _, val = record.partition('=')
+            if key == 'SetUserVar':
+                from base64 import standard_b64decode
+                ukey, has_equal, uval = val.partition('=')
+                self.set_user_var(ukey, (standard_b64decode(uval) if uval else b'') if has_equal == '=' else None)
+
     def desktop_notify(self, osc_code: int, raw_data: str) -> None:
+        if osc_code == 1337:
+            self.osc_1337(raw_data)
         if osc_code == 777:
             if not raw_data.startswith('notify;'):
                 log_error(f'Ignoring unknown OSC 777: {raw_data}')
@@ -932,7 +900,6 @@ class Window:
         if cmd is not None and osc_code == 99:
             self.prev_osc99_cmd = cmd
 
-    # screen callbacks {{{
     def use_utf8(self, on: bool) -> None:
         get_boss().child_monitor.set_iutf8_winid(self.id, on)
 
@@ -1004,7 +971,7 @@ class Window:
             '--ssh-connection-data', json.dumps(conn_data)
         )
 
-    def send_signal_for_key(self, key_num: int) -> bool:
+    def send_signal_for_key(self, key_num: bytes) -> bool:
         try:
             return self.child.send_signal_for_key(key_num)
         except OSError as err:
@@ -1076,18 +1043,6 @@ class Window:
                     tab.relayout_borders()
                 tab.on_bell(self)
 
-    def change_titlebar_color(self) -> None:
-        opts = get_options()
-        val = opts.macos_titlebar_color if is_macos else opts.wayland_titlebar_color
-        if val > 0:
-            if (val & 0xff) == 1:
-                val = self.screen.color_profile.default_bg
-            else:
-                val = val >> 8
-            set_titlebar_color(self.os_window_id, val)
-        else:
-            set_titlebar_color(self.os_window_id, 0, True, -val)
-
     def change_colors(self, changes: Dict[DynamicColor, Optional[str]]) -> None:
         dirtied = default_bg_changed = False
 
@@ -1271,8 +1226,7 @@ class Window:
 
     def handle_remote_print(self, msg: str) -> None:
         text = process_remote_print(msg)
-        print(text, end='', file=sys.stderr)
-        sys.stderr.flush()
+        print(text, end='', flush=True)
 
     def send_cmd_response(self, response: Any) -> None:
         self.screen.send_escape_code_to_child(DCS, '@kitty-cmd' + json.dumps(response))
